# Copyright (C) 2017-2023 elphmod Developers
# This program is free software under the terms of the GNU GPLv3 or later.

"""Step and delta smearing functions."""

from __future__ import division

import numpy as np

try:
    from scipy.special import erf
except ImportError:
    import math

    erf = np.vectorize(math.erf)

xmax = 709.0 # approx. log([max. double] / 2 - 1)

def fermi_dirac(x):
    """Calculate Fermi function."""

    # return 1 - 0.5 * np.tanh(0.5 * x)

    x = np.minimum(x, xmax)

    return 1 / (np.exp(x) + 1)

def fermi_dirac_delta(x):
    """Calculate negative derivative of Fermi function."""

    x = np.minimum(np.absolute(x), xmax)

    return 1 / (2 * np.cosh(x) + 2)

def fermi_dirac_delta_prime(x):
    """Calculate negative 2nd derivative of Fermi function."""

    x = np.sign(x) * np.minimum(np.absolute(x), xmax)

    return -np.sinh(x) / (2 * (np.cosh(x) + 1) ** 2)

def fermi_dirac_entropy(x):
    """Calculate electronic entropy."""

    x = np.sign(x) * np.minimum(np.absolute(x), xmax)

    f = fermi_dirac(x)

    return x * (f - 1) - np.log(f)

fermi_dirac.delta = fermi_dirac_delta
fermi_dirac.delta_prime = fermi_dirac_delta_prime
fermi_dirac.entropy = fermi_dirac_entropy

def gauss(x):
    """Calculate Gaussian step function."""

    return (1 - erf(x)) / 2

def gauss_delta(x):
    """Calculate negative derivative of Gaussian step function."""

    return np.exp(-x * x) / np.sqrt(np.pi)

def gauss_delta_prime(x):
    """Calculate negative 2nd derivative of Gaussian step function."""

    return -2 * x * np.exp(-x * x) / np.sqrt(np.pi)

def gauss_entropy(x):
    """Calculate Gaussian generalized electronic entropy."""

    return gauss_delta(x) / 2

gauss.delta = gauss_delta
gauss.delta_prime = gauss_delta_prime
gauss.entropy = gauss_entropy

def marzari_vanderbilt(x):
    """Calculate Marzari-Vanderbilt (cold smearing) step function."""

    y = x + 1 / np.sqrt(2)

    return (1 - erf(y)) / 2 + np.exp(-y * y) / np.sqrt(2 * np.pi)

def marzari_vanderbilt_delta(x):
    """Calculate negative derivative of Marzari-Vanderbilt step function."""

    y = x + 1 / np.sqrt(2)

    return (np.sqrt(2) * y + 1) * np.exp(-y * y) / np.sqrt(np.pi)

def marzari_vanderbilt_delta_prime(x):
    """Calculate negative 2nd derivative of Marzari-Vanderbilt step function."""

    y = x + 1 / np.sqrt(2)

    return (np.sqrt(2) - 2 * y
        * (np.sqrt(2) * y + 1)) * np.exp(-y * y) / np.sqrt(np.pi)

def marzari_vanderbilt_entropy(x):
    """Calculate Marzari-Vanderbilt generalized electronic entropy."""

    y = x + 1 / np.sqrt(2)

    return y * np.exp(-y * y) / np.sqrt(2 * np.pi)

marzari_vanderbilt.delta = marzari_vanderbilt_delta
marzari_vanderbilt.delta_prime = marzari_vanderbilt_delta_prime
marzari_vanderbilt.entropy = marzari_vanderbilt_entropy

def hermite_polynomials(x, N=100):
    r"""Generate Hermite polynomials.

    .. math::

        H_0(x) &= 1 \\
        H_1(x) &= 2 x \\
        H_{n + 1}(x) &= 2 x H_n(x) - 2 n H_{n - 1}(x)
    """
    H = 0
    h = np.ones_like(x)

    for n in range(N):
        yield h

        h, H = 2 * x * h - 2 * n * H, h

def methfessel_paxton_term(x, order=1, diff=0):
    r"""Calculate Methfessel-Paxton term (without zeroth order).

    From Phys. Rev. B 40, 3616 (1989):

    .. math::

        S_0(x) &= \frac {1 - erf(x)} 2 \\
        S_N(x) &= S_0(x) + \sum_{n = 1}^N A_n H_{2 n - 1}(x) \exp(-x^2) \\
        D_N(x) &= -S'(N, x) = \sum{n = 0}^N A_n H_{2 n}(x) \exp(-x^2) \\
        A_n &= \frac{(-1)^n}{\sqrt \pi n! 4^n}

    This routine has been adapted from Quantum ESPRESSO:

    * Step function: Modules/wgauss.f90
    * Delta function: Modules/w0gauss.f90
    """
    a = 1.0
    s = 0.0
    h = hermite_polynomials(x)

    for n in range(diff):
        a *= -1
        next(h)

    for n in range(1, order + 1):
        a /= -4 * n
        next(h)

        s += a * next(h)

    return s * gauss_delta(x)

def methfessel_paxton(x, N=1):
    """Calculate Methfessel-Paxton step function."""

    return gauss(x) + methfessel_paxton_term(x, order=N, diff=0)

def methfessel_paxton_delta(x, N=1):
    """Calculate negative derivative of Methfessel-Paxton step function."""

    return gauss.delta(x) - methfessel_paxton_term(x, order=N, diff=1)

def methfessel_paxton_delta_prime(x, N=1):
    """Calculate negative 2nd derivative of Methfessel-Paxton step function."""

    return gauss.delta_prime(x) - methfessel_paxton_term(x, order=N, diff=2)

def methfessel_paxton_entropy(x, N=1):
    """Calculate Methfessel-Paxton generalized electronic entropy."""

    return -methfessel_paxton_term(x, order=N, diff=1) / 2

methfessel_paxton.delta = methfessel_paxton_delta
methfessel_paxton.delta_prime = methfessel_paxton_delta_prime
methfessel_paxton.entropy = methfessel_paxton_entropy

def double_fermi_dirac(x, d):
    """Calculate double Fermi function."""

    return (fermi_dirac(x - d) + fermi_dirac(x + d)) / 2

def double_fermi_dirac_delta(x, d):
    """Calculate negative derivative of double Fermi function."""

    return (fermi_dirac.delta(x - d) + fermi_dirac.delta(x + d)) / 2

def double_fermi_dirac_delta_prime(x, d):
    """Calculate negative 2nd derivative of double Fermi function."""

    return (fermi_dirac.delta_prime(x - d) + fermi_dirac.delta_prime(x + d)) / 2

def double_fermi_dirac_entropy(x, d):
    """Calculate double-Fermi-Dirac generalized electronic entropy."""

    return (fermi_dirac.entropy(x - d) + fermi_dirac.entropy(x + d)
        + d * (fermi_dirac(x - d) - fermi_dirac(x + d))) / 2

double_fermi_dirac.delta = double_fermi_dirac_delta
double_fermi_dirac.delta_prime = double_fermi_dirac_delta_prime
double_fermi_dirac.entropy = double_fermi_dirac_entropy

def lorentz(x):
    """Calculate Lorentz step function.

    Used to simulate the influence of a wide box-shaped hybridization function
    at low temperatures. Formula derived by Tim O. Wehling and Erik G.C.P. van
    Loon. Here, we have :math:`x = \epsilon / h` with the height :math:`h` of
    the hybridization, instead of :math:`x = \epsilon / k T` with the
    temperature :math:`T`.
    """
    return 0.5 - np.arctan(x / np.pi) / np.pi

def lorentz_delta(x):
    """Calculate negative derivative of Lorentz step function."""

    return 1.0 / (x * x + np.pi * np.pi)

def lorentz_delta_prime(x):
    """Calculate negative 2nd derivative of Lorentz step function."""

    return -2 * x / (x * x + np.pi * np.pi) ** 2

lorentz.delta = lorentz_delta
lorentz.delta_prime = lorentz_delta_prime

def heaviside(x):
    """Calculate (reflected) Heaviside function."""

    return 0.5 - 0.5 * np.sign(x)

def heaviside_delta(x):
    """Calculate negative derivative of (reflected) Heaviside function."""

    delta = np.copy(x)

    zero = delta == 0

    delta[zero] = np.inf
    delta[~zero] = 0.0

    return delta

def heaviside_delta_prime(x):
    """Calculate negative 2nd derivative of (reflected) Heaviside function."""

    delta_prime = np.copy(x)

    zero = delta_prime == 0

    delta_prime[zero] = -np.copysign(np.inf, delta_prime[zero])
    delta_prime[~zero] = 0.0

    return delta_prime

def heaviside_entropy(x):
    """Calculate negative Heaviside generalized electronic entropy."""

    return np.zeros_like(x)

heaviside.delta = heaviside_delta
heaviside.delta_prime = heaviside_delta_prime
heaviside.entropy = heaviside_entropy

def fermi_dirac_matsubara(x, nmats=1000):
    """Calculate Fermi function as Matsubara sum."""

    inu = 1j * (2 * np.arange(nmats) + 1) * np.pi

    return 0.5 + 2 * np.sum(np.subtract.outer(inu, x) ** -1, axis=0).real

def fermi_dirac_matsubara_delta(x, nmats=1000):
    """Calculate negative derivative of Fermi function as Matsubara sum."""

    inu = 1j * (2 * np.arange(nmats) + 1) * np.pi

    return -2 * np.sum(np.subtract.outer(inu, x) ** -2, axis=0).real

def fermi_dirac_matsubara_delta_prime(x, nmats=1000):
    """Calculate negative 2nd derivative of Fermi function as Matsubara sum."""

    inu = 1j * (2 * np.arange(nmats) + 1) * np.pi

    return -4 * np.sum(np.subtract.outer(inu, x) ** -3, axis=0).real

fermi_dirac_matsubara.delta = fermi_dirac_matsubara_delta
fermi_dirac_matsubara.delta_prime = fermi_dirac_matsubara_delta_prime

def smearing(name='gaussian'):
    """Select smearing function via name used in Quantum ESPRESSO.

    Parameters
    ----------
    name : str, default 'gaussian'
        Any available option for PWscf input parameter ``smearing``.

    Returns
    -------
    function
        Smearing function.
    """
    name = name.lower()

    if name in {'gaussian', 'gauss'}:
        return gauss
    if name in {'methfessel-paxton', 'm-p', 'mp'}:
        return methfessel_paxton
    if name in {'marzari-vanderbilt', 'cold', 'm-v', 'mv'}:
        return marzari_vanderbilt
    if name in {'fermi-dirac', 'f-d', 'fd'}:
        return fermi_dirac

def find_Fermi_level(n, e, kT=0.025, f=fermi_dirac, mu=None, tol=1e-5,
        eps=1e-10):
    """Determine chemical potential via fixed-point iteration.

    See Eqs. 4.21 and 4.22 of https://janber.de/theses/Master_Jan_Berges.pdf.

    Parameters
    ----------
    n : float
        Number of electrons (with spin) per unit cell.
    e : ndarray
        Electronic energies for representative k points.
    kT : float
        Smearing temperature.
    f : function
        Electron distribution as a function of energy divided by `kT`.
    mu : float
        Initial guess for chemical potential. By default, an estimate based on
        the assumption of a constant density of states is used.
    tol : float
        Tolerance for the number of electrons.
    eps : float
        Smallest allowed absolute value of divisor.

    Returns
    -------
    float
        Chemical potential.
    """
    # map number of electrons to whole system and spinless electrons:

    scale = 0.5 * e.size / e.shape[-1]
    n *= scale
    tol *= scale

    # make initial guess for chemical potential:

    if mu is None:
        mu = (e.min() * (e.size - n) + e.max() * n) / e.size

    # solve fixed-point equation:

    while True:
        xi = e - mu
        f0 = f(0.0)
        fx = f(xi / kT)
        w = f0 - fx

        ok = abs(xi) > eps
        w[ok] /= xi[ok]
        w[~ok] = f.delta(0.0) / kT

        if abs(fx.sum() - n) < tol:
            return mu

        mu = (n - e.size * f0 + (e * w).sum()) / w.sum()

def find_Fermi_level_simple(n, e, kT=0.025, f=fermi_dirac, mu=None, tol=1e-5,
        damp=1e-2):
    """Determine chemical potential via simple fixed-point iteration.

    Parameters
    ----------
    n : float
        Number of electrons (with spin) per unit cell.
    e : ndarray
        Electronic energies for representative k points.
    kT : float
        Smearing temperature.
    f : function
        Electron distribution as a function of energy divided by `kT`.
    mu : float
        Initial guess for chemical potential. By default, an estimate based on
        the assumption of a constant density of states is used.
    tol : float
        Tolerance for the number of electrons.
    damp : float
        Damping factor in the fixed-point equation. Large values may prevent
        convergence; small values will slow down convergence.

    Returns
    -------
    float
        Chemical potential.
    """
    # map number of electrons to whole system and spinless electrons:

    scale = 0.5 * e.size / e.shape[-1]
    n *= scale
    tol *= scale

    # make initial guess for chemical potential:

    if mu is None:
        mu = (e.min() * (e.size - n) + e.max() * n) / e.size

    # solve fixed-point equation:

    while True:
        N = f((e - mu) / kT).sum()

        if abs(N - n) < tol:
            return mu

        mu += (n / N - 1) * damp
